from __future__ import annotations

import argparse
import sys
from typing import Any

from . import WorkerConfig, WorkerConfigDict, run

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="worker_manager",
        description="A manager to execute a set of worker processes.",
    )
    parser.add_argument("config", help="configuration file (.py)")
    args = parser.parse_args()

    config_locals: dict[str, Any] = {}
    if args.config == "-":
        code = compile(sys.stdin.read(), "<stdin>", "exec")
        exec(code, config_locals)  # nosec
    else:
        with open(args.config, "rb") as f:
            code = compile(f.read(), args.config, "exec")
            exec(code, config_locals)  # nosec

    if "worker_config" not in config_locals:
        raise RuntimeError(
            "There is no local variable `worker_config` in the configuration file"
        )

    if not isinstance(config_locals["worker_config"], dict):
        raise RuntimeError("The local variable `worker_config` should be a dictionary")

    if not config_locals["worker_config"]:
        raise RuntimeError("The local variable `worker_config` has no entry")

    worker_config_dict: WorkerConfigDict = dict()
    for name, config_raw in config_locals["worker_config"].items():
        worker_config_dict[name] = WorkerConfig(name=name, **config_raw)

    run(worker_config_dict)



from __future__ import annotations

import json
import logging
import logging.handlers
import os
import queue
import random
import sys
import threading
import time
import traceback
import urllib.parse
import urllib.request
from typing import Any

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration

_DEFAULT_LOGGING_FORMAT = (
    "[%(levelname)1.1s %(asctime)s"
    " P%(process)d"
    " %(threadName)s"
    " %(module)s:%(lineno)d]"
    " %(message)s"
)

_MAX_FLUSH_DELAY_BACKOFF = 300  # 300 seconds
_MAX_LOGS_PER_FLUSH = 5


class _SlackLoggerThread(threading.Thread):
    def __init__(self, slack_url: str) -> None:
        super().__init__()
        self.daemon = True
        self.slack_url = slack_url
        self.queue: queue.Queue[dict[str, Any]] = queue.Queue()

        self._hostname = os.getenv("DOCKER_HOSTNAME") or os.getenv("HOSTNAME")
        self._flush_cnt = 0

    def run(self) -> None:
        while True:
            if self.flush() == 0:
                self._flush_cnt = 0
            self._flush_cnt += 1

            # https://cloud.google.com/storage/docs/exponential-backoff
            backoff_time = (2**self._flush_cnt) + (random.randint(0, 1000) / 1000)
            time.sleep(min(backoff_time, _MAX_FLUSH_DELAY_BACKOFF))

    def flush(self) -> int:
        logs: list[dict[str, Any]] = []
        while not self.queue.empty():
            logs.append(self.queue.get())

        if not logs:
            return 0

        msg_text = f"[{len(logs)}] logs from {self._hostname}"
        if _MAX_LOGS_PER_FLUSH < len(logs):
            msg_text += f"\n(truncated to first {_MAX_LOGS_PER_FLUSH} logs)"
            logs = logs[:_MAX_LOGS_PER_FLUSH]

        urllib.request.urlopen(  # nosec: B310
            self.slack_url,
            data=urllib.parse.urlencode(
                {"payload": json.dumps({"text": msg_text, "attachments": logs})}
            ).encode(),
        )

        return len(logs)


class _SlackLoggerHandler(logging.handlers.QueueHandler):
    def __init__(self, product_name: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.product_name = product_name

    def prepare(self, record: logging.LogRecord) -> dict[str, Any]:
        msg_pretext = record.message.splitlines()[0]
        if 80 < len(msg_pretext):
            msg_pretext = msg_pretext[:77] + "..."

        msg_fields = [
            {
                "title": "ProductName",
                "value": self.product_name,
                "short": True,
            },
            {
                "title": "LoggerName",
                "value": record.name,
                "short": True,
            },
            {
                "title": "When",
                "value": record.asctime,
                "short": True,
            },
            {
                "title": "Level",
                "value": record.levelname,
                "short": True,
            },
            {
                "title": "Process",
                "value": f"{record.processName} ({record.process})",
                "short": True,
            },
            {
                "title": "Thread",
                "value": f"{record.threadName} ({record.thread})",
                "short": True,
            },
            {
                "title": "Where",
                "value": (
                    f"{record.pathname}"
                    f" (line {record.lineno}, in {record.funcName})"
                ),
            },
        ]

        msg_text = record.message
        if record.exc_info is not None:
            tb_lines = "".join(
                traceback.format_exception(*record.exc_info)
            ).splitlines()
            if 60 < len(tb_lines):
                tb_lines = [
                    *tb_lines[:30],
                    "...(truncated)...",
                    *tb_lines[-30:],
                ]
            tb_truncated = "\n".join(tb_lines)
            msg_text = f"{msg_text}\n```\n{tb_truncated}\n```"

        return {
            "color": "#ff7777",
            "pretext": msg_pretext,
            "text": msg_text,
            "fields": msg_fields,
        }


def init_logger(
    product_name: str,
    app_logger_name: str,
    app_logger_level: int = logging.INFO,
    stdout_handler_level: int = logging.INFO,
    stdout_handler_format: str = _DEFAULT_LOGGING_FORMAT,
    slack_url: str | None = None,
    slack_handler_level: int = logging.ERROR,
    sentry_dsn: str | None = None,
    sentry_handler_level: int = logging.ERROR,
    additional_loggers: list[logging.Logger] | None = None,
) -> logging.Logger:
    if additional_loggers is None:
        additional_loggers = []

    app_logger = logging.getLogger(app_logger_name)
    app_logger.setLevel(app_logger_level)

    log_handlers: list[logging.Handler] = []

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(stdout_handler_level)
    stdout_handler.setFormatter(logging.Formatter(stdout_handler_format))
    log_handlers.append(stdout_handler)

    if slack_url is not None:
        slack_logger_thread = _SlackLoggerThread(slack_url)
        slack_logger_thread.start()

        slack_handler = _SlackLoggerHandler(product_name, slack_logger_thread.queue)
        slack_handler.setLevel(slack_handler_level)
        log_handlers.append(slack_handler)

    if sentry_dsn is not None:
        # `sentry_sdk` internally uses a thread to send HTTP requests,
        # so we do not need to care.
        sentry_sdk.init(
            sentry_dsn,
            release=product_name,
            integrations=[
                # `LoggingIntegration` is enabled by default,
                # but we define this here fore explicit configuration
                LoggingIntegration(
                    level=app_logger_level,
                    event_level=sentry_handler_level,
                ),
            ],
        )

    interested_loggers = [app_logger, *additional_loggers]

    for logger in interested_loggers:
        for log_hadnler in log_handlers:
            logger.addHandler(log_hadnler)

    app_logger.info("App logger is started.")

    return app_logger



try:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)
except LookupError:
    try:
        from ._version import version

        __version__ = version
    except ModuleNotFoundError:
        raise RuntimeError(
            "Cannot determine version : check git repository or _version.py fle"
        )



from __future__ import annotations

import functools
import os
import signal
import subprocess
import sys
import threading
import time
import traceback
from types import FrameType
from typing import Dict, Iterable, Iterator, NamedTuple, Protocol

_POLL_INTERVAL = 0.1
_SIGNAL_TIMEOUT = 15
_MONITOR_THREAD_TIMEOUT = 5
_CONDITION_CHECK_INTERVAL = 1
_CONDITION_CHECK_TIMEOUT = 10


class WorkerConfig(NamedTuple):
    name: str
    num_workers: int
    cmd: list[str]
    env: dict[str, str] | None = None
    start_condition_cmd: list[str] | None = None
    stop_signal: str = "SIGTERM"
    stop_order: int = 1
    signal_to_pg: bool = False


WorkerConfigDict = Dict[str, WorkerConfig]


class WorkerEntry(NamedTuple):
    worker_config: WorkerConfig
    process_wraper_list: list[ProcessWrapper]


class ProcessWrapper(NamedTuple):
    parent_worker_config: WorkerConfig

    name: str
    popen_obj: subprocess.Popen
    monitor: threading.Thread

    @property
    def pid(self) -> int:
        return self.popen_obj.pid

    @property
    def pgid(self) -> int:
        return os.getpgid(self.pid)


class _PrinterType(Protocol):
    def __call__(self, prefix: str, txt: str, *, is_stdout: bool = False) -> None:
        ...


def _make_process_wrapper_iter(
    worker_entries: Iterable[WorkerEntry],
) -> Iterator[ProcessWrapper]:
    for worker_entry in worker_entries:
        for pwrapper in worker_entry.process_wraper_list:
            yield pwrapper


def _process_launcher(
    term_event: threading.Event,
    printer: _PrinterType,
    worker_entry: WorkerEntry,
) -> None:
    worker_config = worker_entry.worker_config

    if worker_config.start_condition_cmd is not None:
        printer(worker_config.name, "Check start condition...")

        while not term_event.is_set():
            try:
                condition_result = subprocess.run(
                    worker_config.start_condition_cmd,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                    timeout=_CONDITION_CHECK_TIMEOUT,
                )
                if condition_result.returncode == 0:
                    break
            except Exception:
                pass

            printer(worker_config.name, "Start condition is not met")

            time.sleep(_CONDITION_CHECK_INTERVAL)

    for idx in range(worker_config.num_workers):
        if term_event.is_set():
            return

        process_name = f"{worker_config.name}_{idx + 1}"

        popen_obj = subprocess.Popen(
            worker_config.cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=worker_config.env,
            text=True,
            start_new_session=True,
        )

        monitor = threading.Thread(
            target=functools.partial(
                _process_monitor, printer, process_name, popen_obj
            ),
            daemon=True,
        )
        monitor.start()

        pwrapper = ProcessWrapper(
            parent_worker_config=worker_config,
            name=process_name,
            popen_obj=popen_obj,
            monitor=monitor,
        )

        printer(
            process_name,
            f"Started a process (PID {pwrapper.pid}, PGID {pwrapper.pgid})",
        )

        worker_entry.process_wraper_list.append(pwrapper)


def _process_monitor(
    printer: _PrinterType,
    process_name: str,
    popen_obj: subprocess.Popen,
) -> None:
    assert popen_obj.stdout is not None

    while True:
        try:
            output: str = popen_obj.stdout.readline()
        except Exception:
            print(
                'Error while reading outputs from "%s"\n%s'
                % (process_name, traceback.format_exc()),
                flush=True,
            )
            break

        if not output:
            break

        printer(process_name, output.rstrip(), is_stdout=True)

    while True:
        retcode = popen_obj.poll()
        if retcode is not None:
            break
        time.sleep(_POLL_INTERVAL)

    printer(process_name, f"Terminated (retcode: {retcode})")


def run(worker_config_dict: WorkerConfigDict) -> None:
    """Run a set of processes defined by `worker-config`

    .. note:: Example Configuration

        import os

        worker_config = {
            'manager': {
                'num_workers': 1,
                'cmd': ['python', 'launcher_manager.py'],
                'env': {
                    'MANAGER_IN_PORT': '80',
                    'MANAGER_OUT_PORT': '1234',
                },
                'stop_order': 3
            },
            'http-server': {
                'num_workers': 1,
                'cmd': [
                    'gunicorn',
                    '-c', './assets/gunicorn_conf.py',
                    'launcher_http_server:server.app',
                ],
                'stop_order': 2
            },
            'http-server-nginx': {
                'num_workers': 1,
                'cmd': ['nginx', '-g', 'daemon off;'],
                'start_condition_cmd': [
                    'curl',
                    '--fail', '-s',
                    'http://0.0.0.0:8080/ping',
                ],
                'stop_signal': 'SIGQUIT'
            },
            'tcp-server': {
                'num_workers': int(os.environ.get('TCP_SERVER_NUM_WORKERS', 1)),
                'cmd': ['python', 'launcher_tcp_server.py'],
                'stop_order': 2
            }
        }
    """

    term_signal = signal.Signals.SIGTERM
    term_event = threading.Event()

    def _signal_handler(signum: int, frame: FrameType | None) -> None:
        if term_event.is_set():
            print("Forced termination of manager", flush=True)
            sys.exit(1)

        nonlocal term_signal
        term_signal = signal.Signals(signum)
        term_event.set()
        print("Signal handler called with signal :", term_signal.name, flush=True)

    signal.signal(signal.SIGTERM, _signal_handler)
    signal.signal(signal.SIGINT, _signal_handler)

    max_name_length = max(
        len("%s_%d" % (worker_config.name, worker_config.num_workers + 1))
        for worker_config in worker_config_dict.values()
    )

    def _printer(prefix: str, txt: str, *, is_stdout: bool = False) -> None:
        formatstr = "%-{:d}s %s %s".format(max_name_length)
        print(formatstr % (prefix, "|" if is_stdout else ">", txt), flush=True)

    worker_entries: list[WorkerEntry] = []
    for worker_config in worker_config_dict.values():
        if worker_config.stop_signal.upper() not in signal.Signals.__members__:
            raise RuntimeError("Invalid stop signal", worker_config)

        worker_entry = WorkerEntry(worker_config, [])
        worker_entries.append(worker_entry)

        worker_process_launcher = threading.Thread(
            target=functools.partial(
                _process_launcher, term_event, _printer, worker_entry
            ),
            daemon=True,
        )
        worker_process_launcher.start()

    # Monitor the liveness of all worker processes
    any_done_pwrapper: ProcessWrapper | None = None
    try:
        while not term_event.is_set():
            for pwrapper in _make_process_wrapper_iter(worker_entries):
                if pwrapper.popen_obj.poll() is not None:
                    any_done_pwrapper = pwrapper
                    term_event.set()
                    break

                time.sleep(_POLL_INTERVAL)

        else:
            print(
                "Terminate all processes... (cuased by %s)"
                % (any_done_pwrapper.name if any_done_pwrapper else "signal"),
                flush=True,
            )

    except Exception:
        traceback.print_exc()

    stop_order_to_worker_entries: dict[int, list[WorkerEntry]] = {}
    for worker_entry in worker_entries:
        stop_order = worker_entry.worker_config.stop_order
        if stop_order not in stop_order_to_worker_entries:
            stop_order_to_worker_entries[stop_order] = []
        stop_order_to_worker_entries[stop_order].append(worker_entry)

    for _, target_worker_entries in sorted(stop_order_to_worker_entries.items()):
        for pwrapper in _make_process_wrapper_iter(target_worker_entries):
            retcode = pwrapper.popen_obj.poll()
            if retcode is not None:
                continue

            stop_signal = signal.Signals[
                pwrapper.parent_worker_config.stop_signal.upper()
            ]

            _printer(
                pwrapper.name,
                (
                    f"{stop_signal.name} is requested"
                    " (PID: {pwrapper.pid}, PGID: {pwrapper.pgid})"
                ),
            )

            if pwrapper.parent_worker_config.signal_to_pg:
                os.killpg(pwrapper.pgid, stop_signal.value)
            else:
                os.kill(pwrapper.pid, stop_signal.value)

        wait_until_ts = time.monotonic() + _SIGNAL_TIMEOUT
        while time.monotonic() < wait_until_ts:
            for pwrapper in _make_process_wrapper_iter(target_worker_entries):
                retcode = pwrapper.popen_obj.poll()
                if retcode is None:
                    break
            else:
                break
            time.sleep(_POLL_INTERVAL)
        else:
            print(
                "Timeout while waiting the termination of worker processes",
                flush=True,
            )

        wait_until_ts = time.monotonic() + _MONITOR_THREAD_TIMEOUT
        while time.monotonic() < wait_until_ts:
            for pwrapper in _make_process_wrapper_iter(target_worker_entries):
                pwrapper.monitor.join(0)
                if pwrapper.monitor.is_alive():
                    break
            else:
                break
            time.sleep(_POLL_INTERVAL)
        else:
            print(
                "Timeout while waiting the termination of monitor threads",
                flush=True,
            )

    for pwrapper in _make_process_wrapper_iter(worker_entries):
        if pwrapper.monitor.is_alive():
            print(f"Live monitor thread : {pwrapper.name}", flush=True)
